
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

local _G = _G
local InCombatLockdown = InCombatLockdown
local type = type

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database
local M = A:NewModule("Core Registry")

-------------------------------------------------------------------------------
-- Module
-------------------------------------------------------------------------------

function M:OnInitialize()
	if not self:IsHooked("CreateFrame") then
		self:SecureHook("CreateFrame")
	end
end

-------------------------------------------------------------------------------
-- Frame Configuration
-------------------------------------------------------------------------------

local function FrameConfig_Hide(self, frame)
	if not self:IsHooked(frame, "OnShow") then
		self:HookScript(frame, "OnShow", "Frame_OnShow")
	end

	frame:Hide()
end

local function FrameConfig_SetParent(self, frame, parentName)
	frame:SetParent(_G[parentName])
end

local function FrameConfig_ClearAllPoints(self, frame, ...)
	if not self:IsHooked(frame, "ClearAllPoints") then
		self:RawHook(frame, "ClearAllPoints", "TwistingNether", true)
	end

	self.hooks[frame].ClearAllPoints(frame)
end

local function FrameConfig_SetPoint(self, frame, ...)
	if not self:IsHooked(frame, "ClearAllPoints") then
		self:RawHook(frame, "ClearAllPoints", "TwistingNether", true)
		self.hooks[frame].ClearAllPoints(frame)
	end

	if not self:IsHooked(frame, "SetAllPoints") then
		self:RawHook(frame, "SetAllPoints", "TwistingNether", true)
	end

	if not self:IsHooked(frame, "SetPoint") then
		self:RawHook(frame, "SetPoint", "TwistingNether", true)
	end

	self.hooks[frame].SetPoint(frame, ...)
end

local function FrameConfig_SetWidth(self, frame, width)
	if not self:IsHooked(frame, "SetWidth") then
		self:RawHook(frame, "SetWidth", "TwistingNether", true)
	end

	self.hooks[frame].SetWidth(frame, width)
end

local function FrameConfig_SetHeight(self, frame, height)
	if not self:IsHooked(frame, "SetHeight") then
		self:RawHook(frame, "SetHeight", "TwistingNether", true)
	end

	self.hooks[frame].SetHeight(frame, height)
end

local function FrameConfig_SetScale(self, frame, scale)
	frame:SetScale(scale)
end

local function FrameConfig_SetScript(self, frame, handler, func)
	if handler == "OnShow" and frame:IsShown() then
		func(frame)
	end

	if handler == "OnHide" and not frame:IsShown() then
		func(frame)
	end

	frame:SetScript(handler, func)
end

local function FrameConfig_CreateBorder(self, frame, ...)
	B:CreateBorder(frame, ...)
end

local function FrameConfig_SetBackgroundColor(self, frame, ...)
	if frame.GetBorder then
		frame:GetBorder():SetBackgroundColor(...)
	end
end

local function FrameConfig_SetBackgroundStrata(self, frame, ...)
	if frame.GetBorder then
		frame:GetBorder():SetBackgroundStrata(...)
	end
end

local function FrameConfig_SetBorderStrata(self, frame, ...)
	if frame.GetBorder then
		frame:GetBorder():SetBorderStrata(...)
	end
end

local function FrameConfig_Unhook(self, frame)
	self:Unhook(frame, "ClearAllPoints")
	self:Unhook(frame, "SetAllPoints")
	self:Unhook(frame, "SetPoint")
	self:Unhook(frame, "SetWidth")
	self:Unhook(frame, "SetHeight")
	self:Unhook(frame, "SetHeight")
end

-- Register a frame and return a new frame config object.
function M:RegisterFrame(frameName)
	local frameConfig = {}

	local function Run(callback, dependencyName, ...)
		local frame = type(frameName) == "table" and frameName or _G[frameName] 
		local dependency = type(dependencyName) == "table" and dependencyName or _G[dependencyName]

		if not frame then
			-- The frame doesn't exist.
			if type(frameName) == "string" then
				A.FrameQueue:Add(frameName, Run, callback, dependencyName, ...)
			end
		elseif dependencyName and not dependency then
			-- The dependency frame doesn't exist.
			if type(dependencyName) == "string" then
				A.FrameQueue:Add(dependencyName, Run, callback, dependencyName, ...)
			end
		elseif frame:IsProtected() and InCombatLockdown() then
			-- The frame is protected.
			A.CombatQueue:Add(frame, Run, callback, dependencyName, ...)
		else
			callback(self, frame, ...)
		end

		return frameConfig
	end

	function frameConfig:Hide()
		return Run(FrameConfig_Hide)
	end

	function frameConfig:SetParent(parent)
		return Run(FrameConfig_SetParent, parent, parent)
	end

	function frameConfig:ClearAllPoints()
		return Run(FrameConfig_ClearAllPoints)
	end

	function frameConfig:SetPoint(point, relativeTo, relativePoint, offsetX, offsetY)
		return Run(FrameConfig_SetPoint, relativeTo, point, relativeTo, relativePoint, offsetX, offsetY)
	end

	function frameConfig:SetWidth(width)
		return Run(FrameConfig_SetWidth, nil, width)
	end

	function frameConfig:SetHeight(height)
		return Run(FrameConfig_SetHeight, nil, height)
	end

	function frameConfig:SetScale(scale)
		return Run(FrameConfig_SetScale, nil, scale)
	end

	function frameConfig:SetScript(handler, func)
		return Run(FrameConfig_SetScript, nil, handler, func)
	end

	function frameConfig:CreateBorder(...)
		return Run(FrameConfig_CreateBorder, nil, ...)
	end

	function frameConfig:SetBackgroundColor(...)
		return Run(FrameConfig_SetBackgroundColor, nil, ...)
	end

	function frameConfig:SetBackgroundStrata(...)
		return Run(FrameConfig_SetBackgroundStrata, nil, ...)
	end

	function frameConfig:SetBorderStrata(...)
		return Run(FrameConfig_SetBorderStrata, nil, ...)
	end

	function frameConfig:Unhook()
		return Run(FrameConfig_Unhook)
	end

	function frameConfig:AddCallback(callback)
		return Run(callback)
	end

	return frameConfig
end

function A:RegisterFrame(...)
	return M:RegisterFrame(...)
end

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

-- Use a CreateFrame hook to find newly created frames.
function M:CreateFrame(frameType, frameName, parentFrame, inheritsFrame)
	if frameName then
		A.FrameQueue:Process(frameName)
	end
end

-- Hide the frame whenever it is shown.
function M:Frame_OnShow(frame)
	if frame:IsProtected() and InCombatLockdown() then
		A.CombatQueue:Add(frame, self.Frame_OnShow, frame)
	else
		frame:Hide()
	end
end

function M:TwistingNether()
end
